.md-header__button.md-logo :where(img,svg) {
    width: 100%;
    height: 30px;
}

/* .md-nav--primary .md-nav__title[for=__drawer] {
    background-color: var(--md-default-fg-color--lightest);
    color: var(--md-default-fg-color--light);
} */

.md-header, .md-header__inner {
    background-color: #fff;
    color: #2A4CDF;
}

.md-header[data-md-state=shadow] {
    box-shadow: 0px 2px 4px -1px rgb(0 0 0 / 20%), 0px 4px 5px 0px rgb(0 0 0 / 14%), 0px 1px 10px 0px rgb(0 0 0 / 12%);
    transition: box-shadow 200ms linear;
}

.md-header__inner {
    height: 80px;
}

.md-header__topic {
    font-size: 24px;
}

.md-footer {
    background-color: #2A4CDF;
}

.md-search__form:hover {
    background-color: rgba(0,0,0,.13);
}

.md-typeset h1 {
    color: black;
    font-weight: 700;
}

.youtube {
    position: relative;
    width: 100%;
    height: 0;
    padding-bottom: 56.25%;
}

.youtube iframe {
    position: absolute;
    top: 0;
    left: 0;
    width: 100%;
    height: 100%;
}

.home-grid {
    display: grid;
    grid-template-columns: 1fr 1fr;
    grid-column-gap: 24px;
    row-gap: 24px;
}

.home-grid div {
    border-radius: 4px;
    padding: 24px;
    background-color: #f3e9d9;
}

.home-grid h3 {
    margin-top: 0;
    font-weight: 700;
}

.image-container p {
    width: 100%;
    display: flex;
    flex-direction: row;
    flex-wrap: nowrap;
    gap: 1em;
    margin-top: 1em;
    margin-bottom: 1em;
    align-content: stretch
}

.image-container img {
    flex-grow: 1;
    width: 100%;
}

.image-container::after {
    content: "";
    display: table;
    clear: both;
}

.screenshot {
    /* border: 1px solid var(--md-primary-fg-color--dark); */
    box-shadow: var(--md-shadow-z3);
    margin: 1em;
    max-width: calc(100% - 2em) !important;
}

.nomad-button {
    display: inline-block;
    /* font-size: 14px; */
    font-weight: 700;
    text-align: center;
    white-space: nowrap;
    vertical-align: middle;
    cursor: pointer;
    border-radius: 4px !important;
    background-color: #2A4CDF;
    color: #ffffff !important;
    border: none !important;
    margin: 0;

    /* Additional styles from the original SCSS */
    letter-spacing: normal;
    text-transform: none;
  }

.nomad-button--card-action {
    margin: 1em 0 0 0 !important;
}

.click-zoom {
    text-align: center;
}

.click-zoom input[type=checkbox] {
    z-index: 0;
    display: none
}

.click-zoom img {
    transition: transform 0.25s ease;
    z-index: 0;
    cursor: zoom-in
}

.click-zoom input[type=checkbox]:checked~img {
    transform: scale(2.3);
    z-index: 0;
    cursor: zoom-out
}